
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include "spice.h"

SPICE_BEGIN_DECLS
const char *spice_compat_version_t_get_nick(spice_compat_version_t value);
const char *spice_image_compression_t_get_nick(spice_image_compression_t value);
const char *spice_wan_compression_t_get_nick(spice_wan_compression_t value);
SPICE_END_DECLS

/* Generated data ends here */

